<?php
namespace TYPO3\Neos\Controller\Module\Administration;

/*                                                                        *
 * This script belongs to the TYPO3 Flow package "TYPO3.Neos".            *
 *                                                                        *
 * It is free software; you can redistribute it and/or modify it under    *
 * the terms of the GNU General Public License, either version 3 of the   *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The TYPO3 project - inspiring people to share!                         *
 *                                                                        */

use TYPO3\Flow\Annotations as Flow;
use TYPO3\Flow\Error\Error;
use TYPO3\Flow\Error\Message;
use TYPO3\Flow\Error\Warning;
use TYPO3\Flow\Package\Exception\ProtectedPackageKeyException;
use TYPO3\Flow\Package\Exception\UnknownPackageException;
use TYPO3\Flow\Package\Exception;

/**
 * The TYPO3 Package Management module controller
 *
 * @Flow\Scope("singleton")
 */
class PackagesController extends \TYPO3\Neos\Controller\Module\AbstractModuleController {

	/**
	 * @Flow\Inject
	 * @var \TYPO3\Flow\Package\PackageManagerInterface
	 */
	protected $packageManager;

	/**
	 * @return void
	 */
	public function indexAction() {
		$packageGroups = array();
		foreach ($this->packageManager->getAvailablePackages() as $package) {
			/** @var \TYPO3\Flow\Package $package */
			$packagePath = substr($package->getPackagepath(), strlen(FLOW_PATH_PACKAGES));
			$packageGroup = substr($packagePath, 0, strpos($packagePath, '/'));
			$packageGroups[$packageGroup][$package->getPackageKey()] = array(
				'sanitizedPackageKey' => str_replace('.', '', $package->getPackageKey()),
				'version' => $package->getPackageMetaData()->getVersion(),
				'name' => $package->getComposerManifest('name'),
				'type' => $package->getComposerManifest('type'),
				'description' => $package->getPackageMetaData()->getDescription(),
				'metaData' => $package->getPackageMetaData(),
				'isActive' => $this->packageManager->isPackageActive($package->getPackageKey()),
				'isFrozen' => $this->packageManager->isPackageFrozen($package->getPackageKey()),
				'isProtected' => $package->isProtected()
			);
		}
		ksort($packageGroups);
		foreach (array_keys($packageGroups) as $packageGroup) {
			ksort($packageGroups[$packageGroup]);
		}
		$this->view->assignMultiple(array(
			'packageGroups' => $packageGroups,
			'isDevelopmentContext' => $this->objectManager->getContext()->isDevelopment()
		));
	}

	/**
	 * Activate package
	 *
	 * @param string $packageKey Package to activate
	 * @return void
	 */
	public function activateAction($packageKey) {
		$this->flashMessageContainer->addMessage($this->activatePackage($packageKey));
		$this->redirect('index');
	}

	/**
	 * Deactivate package
	 *
	 * @param string $packageKey Package to deactivate
	 * @return void
	 */
	public function deactivateAction($packageKey) {
		$this->flashMessageContainer->addMessage($this->deactivatePackage($packageKey));
		$this->redirect('index');
	}

	/**
	 * Delete package
	 *
	 * @param string $packageKey Package to delete
	 * @return void
	 */
	public function deleteAction($packageKey) {
		$this->flashMessageContainer->addMessage($this->deletePackage($packageKey));
		$this->redirect('index');
	}

	/**
	 * Freeze package
	 *
	 * @param string $packageKey Package to freeze
	 * @return void
	 */
	public function freezeAction($packageKey) {
		$this->flashMessageContainer->addMessage($this->freezePackage($packageKey));
		$this->redirect('index');
	}

	/**
	 * Unfreeze package
	 *
	 * @param string $packageKey Package to freeze
	 * @return void
	 */
	public function unfreezeAction($packageKey) {
		$this->packageManager->unfreezePackage($packageKey);
		$this->flashMessageContainer->addMessage(new Message($packageKey . ' has been unfrozen', 1347464246));
		$this->redirect('index');
	}

	/**
	 * @param array $packageKeys
	 * @param string $action
	 * @return void
	 * @throws \RuntimeException
	 */
	public function batchAction(array $packageKeys, $action) {
		switch ($action) {
			case 'freeze':
				$frozenPackages = array();
				foreach ($packageKeys as $packageKey) {
					$message = $this->freezePackage($packageKey);
					if ($message instanceof Error || $message instanceof Warning) {
						$this->flashMessageContainer->addMessage($message);
					} else {
						array_push($frozenPackages, $packageKey);
					}
				}
				if (count($frozenPackages) > 0) {
					$message = new Message('Following packages have been frozen: ' . implode(', ', $frozenPackages));
				} else {
					$message = new Warning('Unable to freeze the selected packages');
				}
			break;
			case 'unfreeze':
				foreach ($packageKeys as $packageKey) {
					$this->packageManager->unfreezePackage($packageKey);
				}
				$message = new Message('Following packages have been unfrozen: ' . implode(', ', $packageKeys));
			break;
			case 'activate':
				$activatedPackages = array();
				foreach ($packageKeys as $packageKey) {
					$message = $this->activatePackage($packageKey);
					if ($message instanceof Error || $message instanceof Warning) {
						$this->flashMessageContainer->addMessage($message);
					} else {
						array_push($activatedPackages, $packageKey);
					}
				}
				if (count($activatedPackages) > 0) {
					$message = new Message('Following packages have been activated: ' . implode(', ', $activatedPackages));
				} else {
					$message = new Warning('Unable to activate the selected packages');
				}
			break;
			case 'deactivate':
				$deactivatedPackages = array();
				foreach ($packageKeys as $packageKey) {
					$message = $this->deactivatePackage($packageKey);
					if ($message instanceof Error || $message instanceof Warning) {
						$this->flashMessageContainer->addMessage($message);
					} else {
						array_push($deactivatedPackages, $packageKey);
					}
				}
				if (count($deactivatedPackages) > 0) {
					$message = new Message('Following packages have been deactivated: ' . implode(', ', $deactivatedPackages));
				} else {
					$message = new Warning('Unable to deactivate the selected packages');
				}
			break;
			case 'delete':
				$deletedPackages = array();
				foreach ($packageKeys as $packageKey) {
					$message = $this->deletePackage($packageKey);
					if ($message instanceof Error || $message instanceof Warning) {
						$this->flashMessageContainer->addMessage($message);
					} else {
						array_push($deletedPackages, $packageKey);
					}
				}
				if (count($deletedPackages) > 0) {
					$message = new Message('Following packages have been deleted: ' . implode(', ', $deletedPackages));
				} else {
					$message = new Warning('Unable to delete the selected packages');
				}
			break;
			default:
				throw new \RuntimeException('Invalid action "' . $action . '" given.', 1347463918);
		}

		$this->flashMessageContainer->addMessage($message);
		$this->redirect('index');
	}

	/**
	 * @param string $packageKey
	 * @return Error|Message
	 */
	protected function activatePackage($packageKey) {
		try {
			$this->packageManager->activatePackage($packageKey);
			$message = new Message('The package ' . $packageKey . ' is activated', 1343231680);
		} catch (UnknownPackageException $exception) {
			$message = new Error('The package ' . $packageKey . ' is not present and can not be activated', 1343231681);
		}
		return $message;
	}

	/**
	 * @param string $packageKey
	 * @return Error|Message
	 */
	protected function deactivatePackage($packageKey) {
		try {
			$this->packageManager->deactivatePackage($packageKey);
			$message = new Message($packageKey . ' was deactivated', 1343231678);
		} catch (ProtectedPackageKeyException $exception) {
			$message = new Error('The package ' . $packageKey . ' is protected and can not be deactivated', 1343231679);
		}
		return $message;
	}

	/**
	 * @param string $packageKey
	 * @return Error|Message
	 */
	protected function deletePackage($packageKey) {
		try {
			$this->packageManager->deletePackage($packageKey);
			$message = new Message($packageKey . ' has been deleted', 1343231685);
		} catch (UnknownPackageException $exception) {
			$message = new Error($exception->getMessage(), 1343231686);
		} catch (ProtectedPackageKeyException $exception) {
			$message = new Error($exception->getMessage(), 1343231687);
		} catch (Exception $exception) {
			$message = new Error($exception->getMessage(), 1343231688);
		}
		return $message;
	}

	/**
	 * @param string $packageKey
	 * @return Error|Message
	 */
	protected function freezePackage($packageKey) {
		try {
			$this->packageManager->freezePackage($packageKey);
			$message = new Message($packageKey . ' has been frozen', 1343231689);
		} catch (\LogicException $exception) {
			$message = new Error($exception->getMessage(), 1343231690);
		} catch (UnknownPackageException $exception) {
			$message = new Error($exception->getMessage(), 1343231691);
		}
		return $message;
	}

}
